from . import __about__

from functools import cached_property
from pathlib import Path

import polars as pl

from . import datatypes as _types
from . import validate as _validate
from .parameters import rcParams
from .uniprot import fasta as _fasta

__version__ = __about__.__version__

class Study:
    """
    A FLiPPR Study is used to coordinate Limited Proteolysis (LiP) data processing for a single experiment analyzed with FragPipe.
    Studies can contain LiP data alone or LiP and TrP data; with the inclusion of a FASTA file for automatic metadata generation.
    TrP data should result from the same experimental conditions as the LiP data, but without including a broad-spectrum protease like Proteinase K (PK).

    Args:
        lip (str | Path): Directory or path to the FragPipe LFQ output data for Limited Proteolysis (LiP) experiment data.
        trp (str | Path, optional): Directory or path to the FragPipe LFQ output data for Trypsin Only (TrP) experiment data. Defaults to None.
        fasta (str | Path, optional): Path to the FASTA file containing the same protein sequences used in the LiP/TrP experiments. Defaults to None.
        method (str): Method used to quantify ion intensities in FragPipe. `lfq` - Label-Free Quantification, `silac` - Stable Isotope Labelling (currently not incorporated). Defaults to `lfq`.

    Examples:
        Analysis of a single dataset
        >>> flippr.Study(lip = "PXD025926/FragPipe/LiP_LFQ")

        Analysis of a single dataset with metadata integration
        >>> flippr.Study(lip = "PXD025926/FragPipe/LiP_LFQ", fasta = "UP000000625.fasta")

        Including protein-level normalization from trypsin-only data
        >>> flippr.Study(lip = "PXD025926/FragPipe/LiP_LFQ", trp = "PXD025926/FragPipe/TrP_LFQ")

    """

    def __init__(self, lip: str | Path, trp: str | Path | None = None, fasta: str | Path | None = None, method: str = "lfq") -> None:

        _lip, _trp, _fasta_path, _method = _validate._validate_study(lip, trp, fasta, method)

        self.lip: Path = _lip
        self.trp: Path | None = _trp
        self._fasta_path: Path | None = _fasta_path
        self.method: str = _method

        self._pid: int = 0
        self.processes: dict = dict()
        self.results: dict = dict()

    @property
    def samples(self) -> dict:
        """
        Returns the names of the samples or experimental annotations found in the FragPipe outputs for LiP and TrP (if include) datasets.

        Examples:
            Including only LiP data
            >>> e_coli_refolding = flippr.Study(lip = "PXD025926/FragPipe/LiP_LFQ")
            >>> print(e_coli_refolding.samples)
            {"LiP": ("Native", "Refolded_01_min", "Refolded_05_min", "Refolded_30_min")}

            Including LiP and TrP data
            >>> e_coli_refolding = flippr.Study(lip = "PXD025926/FragPipe/LiP_LFQ", trp = "PXD025926/FragPipe/TrP_LFQ")
            >>> print(e_coli_refolding.samples)
            {"LiP": ("Native", "Refolded_01_min", "Refolded_05_min", "Refolded_30_min"),
             "TrP": ("Native", "Refolded")}

        """

        if self.trp is not None:
            return {"LiP": self.lip_samples, "TrP": self.trp_samples}

        return {"LiP": self.lip_samples}

    @property
    def lip_samples(self) -> set:
        """
        Returns the names of the samples or experimental annotations found in the FragPipe outputs for LiP datasets.

        """

        return self.__read_annotation(self.lip)

    @property
    def trp_samples(self) -> set:
        """
        Returns the names of the samples or experimental annotations found in the FragPipe outputs for TrP datasets.

        """

        return self.__read_annotation(self.trp)

    @cached_property
    def fasta(self) -> pl.DataFrame | None:
        """
        Returns a three column dataframe containing Protein IDs, Sequences, and Valid ID if the Protein ID is interpretable as an official Uniprot ID.

        """
        if self._fasta_path is not None:
            return _fasta._read_fasta(self._fasta_path)

        return None

    def add_process(
        self, pid: object | None, lip_ctrl: str, lip_test: str, n_rep: int, trp_ctrl: str | None = None, trp_test: str | None = None, trp_n_rep: int | None = None) -> None:
        """
        Adding a process to calculate the fold-change between test and control conditions.
        Normalization is optional and must include a TrP experiment in the `Study` set-up.

        Args:
            pid (Any): Process ID used to index the `Study().results` dictionary after running `Study().run()`.
            lip_ctrl (str): Control condition sample name from the LiP experiment.
            lip_test (str): Test condition sample name from the LiP experiment.
            n_rep (int): Number of replicates in the LiP experiment.
            trp_ctrl (str, optional): Control condition sample name from the TrP experiment.
            trp_test (str, optional): Test condition sample name from the TrP experiment.
            trp_n_rep (int, optional): Number of replicates in the TrP experiment.

        Examples:
            Simple fold-change calculation between two conditions
            >>> study.add_process("sample", "WT", "Drug", 3)

            Fold-change calculation with normalization
            >>> study.add_process("sample", "WT", "Drug", 3, "WT_TrP", "DMSO_TrP", 3)

            Add multiple processes to the same study
            >>> study.add_process("cond_1", "WT", "Cond1", 5)
            >>> study.add_process("cond_2", "WT", "Cond2", 5, "WT_TrP", "DMSO_TrP", 5)
        """

        if pid is None:
            pid = self._pid
            self._pid += 1

        self.processes.update(
            {
                pid: _types.Process(
                    self.fasta,
                    self.lip,
                    self.trp,
                    pid,
                    lip_ctrl,
                    lip_test,
                    n_rep,
                    trp_ctrl,
                    trp_test,
                    trp_n_rep,
                )
            }
        )

    def run(self) -> dict[object: _types.Result]:
        """
        Run the processes added to the study.
        Study parameters can be changed by editing the `flippr.rcParams` dictionary.
        """

        self.results = {pid: proc.run(rcParams) for pid, proc in self.processes.items()}

        return self.results

    def __read_annotation(self, liptrp: Path) -> set:
        def __strip_fragpipe_repilicate(sample: str) -> str:
            return "_".join([_ for _ in sample.split("_")[:-1]])

        annotation = liptrp.joinpath("experiment_annotation.tsv")

        annotation_df = pl.read_csv(annotation, separator="\t")

        annotation_dict = annotation_df.select(pl.col("sample")).to_dict(
            as_series=False
        )

        return set(__strip_fragpipe_repilicate(sample) for sample in annotation_dict["sample"])



from pathlib import Path

from .parameters import _LFQ_FP_FILES

# from flippr.parameters import SILAC_FP_FILES


def _validate_study(
    lip: str | Path, trp: str | Path | None, fasta: str | Path | None, method: str
) -> tuple[Path, Path | None, Path | None, str]:
    method = __validate_method(method)

    lip = __validate_fragpipe_path(lip, "lip", method)

    if trp is not None:
        trp = __validate_fragpipe_path(trp, "trp", method)

    if fasta is not None:
        fasta = __validate_fasta_file(fasta)

    return lip, trp, fasta, method


def __validate_method(method: str) -> str:
    """
    Validate the quantification method.
    """

    if not isinstance(method, str):
        raise TypeError(
            f'`method` was provided: "{method}" with type `{type(method)}`. The type `{type(method)}`is not recognized. Set `method` to "lfq" or "silac".'
        )

    if method not in ["lfq", "silac"]:
        raise ValueError(
            f'`method` was provided: "{method}". "{method}" is not recognized. Set `method` to "lfq" or "silac".'
        )

    return method


def __validate_fragpipe_path(path: str | Path, liptrp: str, method: str) -> Path:
    """
    Validate FragPipe output directory paths.
    """

    if not isinstance(path, (str, Path)):
        raise TypeError(
            f'`{liptrp}` was provided: "{path}" with type `{type(path)}`. The type `{type(path)}` is not recognized. Set `{liptrp}` to a FragPipe output path.'
        )

    # type cast to `Path`
    # this should have no effect on `Path`
    path = Path(path)

    # invokes `.exists` and checks for directory-ness
    if not path.is_dir():
        raise ValueError(
            f'`{liptrp}` was provided: "{path}". "{path}" is not a directory path. Set `{liptrp}` to a FragPipe output directory path.'
        )

    __validate_fragpipe_files(path, method)

    return Path(path)


def __validate_fragpipe_files(path: Path, method: str) -> None:
    """
    Validate FragPipe output directory path contains all the necessay files.
    """

    if method == "lfq":
        if not all([path.joinpath(f).exists() for f in _LFQ_FP_FILES]):
            raise FileNotFoundError(
                f'Files not found in "{path}". The FragPipe output directory path should minimally contain: '
                + "\t\n".join([f"`{f}`" for f in _LFQ_FP_FILES])
            )

    if method == "silac":
        raise ValueError("This feature does not work yet")
        # if not all([path.joinpath(f).exists() for f in SILAC_FP_FILES]):
        #     raise FileNotFoundError(f"Files not found in \"{path}\". The FragPipe output directory path should minimally contain: " + "\t\n".join([f"`{f}`" for f in SILAC_FP_FILES]))


def __validate_fasta_file(fasta: str | Path) -> Path:
    """
    Validate Uniprot FASTA file.
    """

    if not isinstance(fasta, (str, Path)):
        raise TypeError(
            f'`fasta` was provided: "{fasta}" with type `{type(fasta)}`. The type `{type(fasta)}` is not recognized. Set `fasta` to a Uniprot FASTA file.'
        )

    # type cast to `Path`
    # this should have no effect on `Path`
    fasta = Path(fasta)

    if not fasta.exists():
        raise FileNotFoundError(
            f'"{fasta.name}" was not found in {fasta.parent.resolve()}. Check that your file extension is `.fasta` or `.faa`. Set `fasta` to a Uniprot FASTA file.'
        )

    if fasta.suffix.lower() not in [".fasta", ".faa"]:
        raise ValueError(
            f'"{fasta.name}" has extension `{fasta.suffix}`. Check that your file extension is `.fasta` or `.faa`. Set `fasta` to a Uniprot FASTA file.'
        )

    return Path(fasta)



import re
from pathlib import Path

import polars as pl
from Bio.SeqIO.FastaIO import SimpleFastaParser

from flippr.parameters import _STANDARD_AA_CONVERSION

# UniProtKB accession number validation regex (https://www.uniprot.org/help/accession_numbers)
ACCESSION_RE = re.compile(
    r"[OPQ][0-9][A-Z0-9]{3}[0-9]|[A-NR-Z][0-9]([A-Z][A-Z0-9]{2}[0-9]){1,2}"
)


def _read_fasta(fasta_path: str | Path) -> pl.DataFrame:
    """
    This function reads and parses Uniprot FASTA files and returns a Polars dataframe.

    Args:
        fasta_path: Path to a Uniprot FASTA file.

    Returns:
        fasta_df: Polars dataframe containing Uniprot FASTA entry data.

    Raises:
        FileNotFoundError: If the `fasta_path` does not exsist.

        ValueError: If the `fasta_path` does not point to a file with `.fasta` or `.faa` extensions.
    """

    fasta_dict = dict()
    with open(fasta_path) as open_fasta:
        fasta_gen = SimpleFastaParser(open_fasta)

        for faa in fasta_gen:
            acc = __extract_accession(faa[0])
            seq = faa[1].upper()

            for old_aa, new_aa in _STANDARD_AA_CONVERSION.items():
                seq = seq.replace(old_aa, new_aa)

            fasta_dict.update({acc: seq})

    fasta_df = pl.from_dict(
        {"Protein ID": list(fasta_dict.keys()), "Sequence": list(fasta_dict.values())}
    )

    fasta_df = fasta_df.with_columns(
        pl.col("Protein ID").map_elements(__validate_accession, return_dtype=pl.Boolean).alias("Valid ID")
    )

    return fasta_df


def __extract_accession(fasta_header: str) -> str:
    """
    This function extracts the accession number from a Uniprot FASTA header line.

    e.g.
    The Uniprot FASTA header...
        >sp|P08200|IDH_ECOLI Isocitrate dehydrogenase
    produced the accession number...
        P08200

    Most other FASTA headers will return the whole header string.

    Args:
        fasta_header: First line from a two-line Uniprot FASTA entry.

    Returns:
        Accession number string | FASTA header string.
    """

    if "|" not in fasta_header:
        return fasta_header

    try:
        return fasta_header.split("|")[1]

    except IndexError:
        return fasta_header


def __validate_accession(accession: str) -> bool:
    """
    This function will return `True` if an accession number fully matches the Uniprot accession number regex.

    For more information, visit https://www.uniprot.org/help/accession_numbers

    Args:
        accession: Accession number string | FASTA header string.

    Returns:
        True | False
    """

    if re.fullmatch(ACCESSION_RE, accession) is not None:
        return True

    return False



from __future__ import annotations

from functools import cached_property
from pathlib import Path

import polars as pl

from . import combine as _combine
from . import functions as _functions
from .parameters import (
    _FLIPPR_PROTEIN_SUMMARY_COLUMNS,
    _LFQ_FP_CONSTANT_ION_COLUMNS,
    _LFQ_FP_CONSTANT_PROTEIN_COLUMNS,
    _LFQ_FP_VARIABLE_ION_COLUMNS,
    _LFQ_FP_VARIABLE_PROTEIN_COLUMNS,
)
from .uniprot import metadata as _metadata


class Process:
    """Organizes individual process runs"""

    def __init__(
        self,
        fasta: pl.DataFrame,
        lip_path: Path,
        trp_path: Path | None,
        pid: object | None,
        lip_ctrl: str,
        lip_test: str,
        n_rep: int,
        trp_ctrl: str | None = None,
        trp_test: str | None = None,
        trp_n_rep: int | None = None,
    ) -> None:
        """docstring"""

        self._is_trp_norm = all(
            trp is not None for trp in [trp_path, trp_ctrl, trp_test, trp_n_rep]
        )

        self._fasta: pl.DataFrame | None = fasta
        self._pid: object = pid

        self._lip_path: Path = lip_path
        self._lip_ctrl_name: str = lip_ctrl
        self._lip_test_name: str = lip_test
        self._lip_ctrl: list[str] = [f"{lip_ctrl}_{n+1}" for n in range(n_rep)]
        self._lip_test: list[str] = [f"{lip_test}_{n+1}" for n in range(n_rep)]
        self._n_rep: int = n_rep
        self._trp_path: Path | None = None
        self._trp_ctrl_name: str | None = None
        self._trp_test_name: str | None = None
        self._trp_ctrl: list[str] | None = None
        self._trp_test: list[str] | None = None
        self._trp_n_rep: int | None = None

        if self._is_trp_norm:
            self._trp_path = trp_path
            self._trp_ctrl_name = trp_ctrl
            self._trp_test_name = trp_test
            self._trp_ctrl = [f"{trp_ctrl}_{n+1}" for n in range(trp_n_rep)]
            self._trp_test = [f"{trp_test}_{n+1}" for n in range(trp_n_rep)]
            self._trp_n_rep = trp_n_rep

    @property
    def name(self) -> str:
        """
            Returns the name of the process from the given LiP control and test conditions.
        """
        return f"{self._lip_ctrl_name}_v_{self._lip_test_name}"

    def run(self, rcParams: dict):
        result = Result(self, rcParams)
        return result


class Result:
    """Organizes a FLiPPR Result"""

    def __init__(self, cls: Process, rcParams: dict) -> None:
        """doctstring"""

        self.rcParams: dict = rcParams
        self._fasta = cls._fasta

        self.name: str = cls.name
        self._is_trp_norm: bool = cls._is_trp_norm

        self._fc = "FC"
        self._n_rep = cls._n_rep
        self._lip_path: Path = cls._lip_path
        self._trp_path: Path = cls._trp_path
        self._lip_ctrl_name: str = cls._lip_ctrl_name
        self._lip_test_name: str = cls._lip_test_name
        self._lip_ctrl_ints: list[str] = [f"{rep} Intensity" for rep in cls._lip_ctrl]
        self._lip_test_ints: list[str] = [f"{rep} Intensity" for rep in cls._lip_test]
        self._trp_n_rep: int | None = None
        self._trp_ctrl_name: str | None = None
        self._trp_test_name: str | None = None
        self._trp_ctrl_ints: list[str] | None = None
        self._trp_test_ints: list[str] | None = None
        self._trp_ctrl_cols: list[str] | None = None
        self._trp_test_cols: list[str] | None = None
        self._trp_process_cols: list[str] | None = None
        self._norm_factors: pl.DataFrame | None = None

        self._lip_ctrl_cols: list[str] = [
            f"{rep} {var}"
            for var in _LFQ_FP_VARIABLE_ION_COLUMNS
            for rep in cls._lip_ctrl
        ]

        self._lip_test_cols: list[str]  = [
            f"{rep} {var}"
            for var in _LFQ_FP_VARIABLE_ION_COLUMNS
            for rep in cls._lip_test
        ]

        self._lip_process_cols: list[str]  = (
            _LFQ_FP_CONSTANT_ION_COLUMNS
            + self._lip_ctrl_cols
            + self._lip_test_cols
        )

        args = {"ctrl_name":self._lip_ctrl_name, "test_name":self._lip_test_name,
                "ctrl_ints":self._lip_ctrl_ints, "test_ints": self._lip_test_ints,
                "n_rep": self._n_rep,
                "rcParams": self.rcParams}

        self._ion = self._read_fragpipe_tsv(self._lip_path, self._lip_process_cols, "combined_ion.tsv")
        self._ion = self.run(self._ion, args)
        self._ion = self.clean_up(self._ion, args)

        if self._is_trp_norm:
            trp_prot_int_val = self.rcParams.get("trp_protein.intensity_value", "MaxLFQ Intensity")

            self._trp_n_rep = cls._trp_n_rep
            self._trp_ctrl_name = cls._trp_ctrl_name
            self._trp_test_name = cls._trp_test_name
            self._trp_ctrl_ints = [f"{rep} {trp_prot_int_val}" for rep in cls._trp_ctrl]
            self._trp_test_ints = [f"{rep} {trp_prot_int_val}" for rep in cls._trp_test]

            self._trp_ctrl_cols = [
                f"{rep} {var}"
                for var in _LFQ_FP_VARIABLE_PROTEIN_COLUMNS
                for rep in cls._trp_ctrl
            ]

            self._trp_test_cols = [
                f"{rep} {var}"
                for var in _LFQ_FP_VARIABLE_PROTEIN_COLUMNS
                for rep in cls._trp_test
            ]

            self._trp_process_cols = (
                _LFQ_FP_CONSTANT_PROTEIN_COLUMNS
                + self._trp_ctrl_cols
                + self._trp_test_cols
            )

            trp_args = {"ctrl_name":self._trp_ctrl_name, "test_name":self._trp_test_name,
                        "ctrl_ints":self._trp_ctrl_ints, "test_ints": self._trp_test_ints,
                        "n_rep": self._trp_n_rep,
                        "rcParams": self.rcParams}

            self._trp_norm = self._read_fragpipe_tsv(self._trp_path, self._trp_process_cols, "combined_protein.tsv")
            self._trp_norm = self.run(self._trp_norm, trp_args)
            self._ion = _functions._normalize_ratios(self._ion, self._trp_norm, rcParams)
            self._fc = "Normalized FC" # Generated after running `._normalize_ratios()`


    def run(self, df: pl.DataFrame, args: dict) -> pl.DataFrame:
        # Can be performed on ion, mod_pep, pep, or protein
        df = _functions._cull_intensities(df, **args)
        df = _functions._add_alt_hypothesis(df, **args)
        df = _functions._impute_aon_intensities(df, **args)
        df = _functions._add_ttest(df, **args)
        df = _functions._add_fdr(df, **args)
        df = _functions._add_ratio(df, **args)
        df = _functions._log2(df, self._fc)
        df = _functions._log10(df, "P-value")
        df = _functions._log10(df, "Adj. P-value")

        return df

    def clean_up(self, df: pl.DataFrame, args: dict) -> pl.DataFrame:
        # Only meant to be performed on the lip ions
        df = _functions._add_start_end_aa(df, **args)
        df = _functions._add_half_trpytic(df, **args)
        df = _functions._add_cut_sites(df, **args)

        return df

    def _read_fragpipe_tsv(self, path: Path, cols: list[str], filename: str):
        df: pl.DataFrame = pl.read_csv(path.joinpath(filename), separator="\t").select(pl.col(cols))
        return df

    def _add_metadata(self, df: pl.DataFrame) -> pl.DataFrame:
        if self._fasta is not None:
            md = _metadata.bio_metadata(self._fasta)
            return df.join(md, on="Protein ID")


    @property
    def ion(self) -> pl.DataFrame:
        """
            ion dataframe
        """
        return self._ion
    
    @property
    def trp_protein(self) -> pl.DataFrame:
        return self._trp_norm

    @cached_property
    def modified_peptide(self) -> pl.DataFrame:
        return _combine.combine_by(self.ion, by="MODIFIED PEPTIDE", fc=self._fc)

    @cached_property
    def peptide(self) -> pl.DataFrame:
        return _combine.combine_by(self.ion, by="PEPTIDE", fc=self._fc)

    @cached_property
    def cut_site(self) -> pl.DataFrame:
        return _combine.combine_by(self.ion, by="CUT SITE", fc=self._fc)

    @cached_property
    def protein_summary(self) -> pl.DataFrame:
        self._proteins = self.ion.group_by("Protein ID", maintain_order=True).agg(
            pl.col(_FLIPPR_PROTEIN_SUMMARY_COLUMNS).first()
        )

        __mod = _combine.summary_by(self.modified_peptide, by="Modified Peptides", fc=self._fc, rcParams=self.rcParams)

        __pep = _combine.summary_by(self.peptide, by="Peptides", fc=self._fc, rcParams=self.rcParams)

        __cut = _combine.summary_by(self.cut_site, by="Cut Sites", fc=self._fc, rcParams=self.rcParams)

        self._proteins = (
            self._proteins
            .join(__mod, on="Protein ID")
            .join(__pep, on="Protein ID")
            .join(__cut, on="Protein ID")
        )

        if self._fasta is not None:
            return self._add_metadata(self._proteins)
        else:
            return self._proteins



import numpy as np
import polars as pl
import scipy as sp

from .functions import _log2, _log10
from .parameters import _FLIPPR_COMBINE_KEY

COMB_NAME_COLUMN = {
    "CUT SITE": "Cut Site ID",
    "PEPTIDE": "Peptide Sequence",
    "MODIFIED PEPTIDE": "Modified Sequence",
}

def combine_by(df: pl.DataFrame, by: str, fc: str) -> pl.DataFrame:

    combined = (
        df.group_by(["Protein ID", COMB_NAME_COLUMN[by]], maintain_order=True)
        .agg(
            pl.col(_FLIPPR_COMBINE_KEY[by]).first(),
            pl.col(["P-value", "Adj. P-value", "CV", fc])
            .filter(
                pl.col("T-test").sign() == pl.col("T-test").sign().sum().sign()
            )
        )
        .with_columns(
            pl.when(pl.col("P-value").list.len().gt(0))
            .then(pl.col("P-value"))
            .otherwise([1.0])
            .alias("P-value"),
            pl.when(pl.col("Adj. P-value").list.len().gt(0))
            .then(pl.col("Adj. P-value"))
            .otherwise([1.0])
            .alias("Adj. P-value"),
            pl.when(pl.col("CV").list.len().gt(0))
            .then(pl.col("CV"))
            .otherwise([0.0])
            .alias("CV"),
            pl.when(pl.col(fc).list.len().gt(0))
            .then(pl.col(fc))
            .otherwise([0.0])
            .alias(fc)
        )
        .with_columns(
            pl.col("P-value")
            .map_elements(lambda x: sp.stats.combine_pvalues(x)[1], return_dtype=pl.Float64)
            .alias("P-value"),
            pl.col("Adj. P-value")
            .map_elements(lambda x: sp.stats.combine_pvalues(x)[1], return_dtype=pl.Float64)
            .alias("Adj. P-value"),
            pl.col("CV").list.max()
            .alias("CV"),
            pl.col(fc).list.median()
            .alias(fc)
        )
    )

    combined = _log2(combined, fc)
    combined = _log10(combined, "P-value")
    combined = _log10(combined, "Adj. P-value")

    return combined


def summary_by(df: pl.DataFrame, by: str, fc: str, rcParams: dict) -> pl.DataFrame:

    prot_fc_sig = rcParams.get("protein.fc_sig_thresh", 1.0)
    prot_pv_sig = rcParams.get("protein.pval_sig_thresh", 0.01)
    prot_apv_sig = rcParams.get("protein.adj_pval_sig_thresh", 0.05)
    prot_hfc_sig = rcParams.get("protein.high_fc_sig_thresh", 6.0)
    prot_hpv_sig = rcParams.get("protein.high_pval_sig_thresh", 0.016)

    val = (
        df.group_by("Protein ID", maintain_order=True)
        .agg(pl.col("CV").filter(pl.col("-Log10 P-value").gt(0)).len())
        .rename({"CV": f"No. of Valid {by}"})
    )

    sig = (
        df.group_by("Protein ID", maintain_order=True)
        .agg(
            pl.col("CV")
            .filter(
                (
                    (pl.col(f"Log2 {fc}").abs().ge(prot_fc_sig))
                    & (pl.col("P-value").le(prot_pv_sig))
                )
                | ( # Legacy, maybe we should drop this or improve it
                    (pl.col(f"Log2 {fc}").abs().ge(prot_hfc_sig))
                    & (pl.col("P-value").le(prot_hpv_sig))
                )
            )
            .len()
        )
        .rename({"CV": f"No. of Significant {by} (P-value)"})
    )

    sigsig = (
        df.group_by("Protein ID", maintain_order=True)
        .agg(
            pl.col("CV")
            .filter(

                    (pl.col(f"Log2 {fc}").abs().ge(prot_fc_sig))
                    & (pl.col("Adj. P-value").le(prot_apv_sig))

            )
            .len()
        )
        .rename({"CV": f"No. of Significant {by} (Adj. P-value)"})
    )

    return val.join(sig, on="Protein ID").join(sigsig, on="Protein ID")



import numpy as np
import polars as pl
import scipy as sp


def _cull_intensities(df: pl.DataFrame,
                      ctrl_name: str,  test_name: str,
                      ctrl_ints: list[str],  test_ints: list[str],
                      n_rep: int,
                      rcParams: dict, **kwargs) -> pl.DataFrame:

    max_missing = rcParams.get("ion.missing_intensity_thresh", 1)

    df = \
    df.with_columns(# Count the number of zeros intensity replicates
        pl.concat_list(ctrl_ints).list.count_matches(0).alias(f"{ctrl_name} ZC"),
        pl.concat_list(test_ints).list.count_matches(0).alias(f"{test_name} ZC"),
    ).filter(# Cull based on zero count (ZC)
        (pl.col(f"{ctrl_name} ZC").le(max_missing) & pl.col(f"{test_name} ZC").eq(0))
        | (pl.col(f"{ctrl_name} ZC").eq(0)           & pl.col(f"{test_name} ZC").le(max_missing))
        | (pl.col(f"{ctrl_name} ZC").eq(n_rep)       & pl.col(f"{test_name} ZC").eq(0))
        | (pl.col(f"{ctrl_name} ZC").eq(0)           & pl.col(f"{test_name} ZC").eq(n_rep))
    ).with_columns(# Replace zero values remaining with null
        pl.col(ctrl_ints).replace(0.0, None),
        pl.col(test_ints).replace(0.0, None)
    )

    return df


def _add_alt_hypothesis(df: pl.DataFrame,
                        ctrl_name: str,  test_name: str,
                        n_rep: int,
                        rcParams: dict, **kwargs) -> pl.DataFrame:

    df = \
    df.with_columns(# Add alternative hypothesis
        pl.when(pl.col(f"{ctrl_name} ZC").eq(n_rep) & pl.col(f"{test_name} ZC").eq(0))
        .then(pl.lit("less"))
        .when(pl.col(f"{ctrl_name} ZC").eq(0) & pl.col(f"{test_name} ZC").eq(n_rep))
        .then(pl.lit("greater"))
        .otherwise(pl.lit("two-sided"))
        .alias("Alternative Hypothesis")
    )

    return df


def _impute_aon_intensities(df: pl.DataFrame,
                            ctrl_name: str,  test_name: str,
                            ctrl_ints: list[str],  test_ints: list[str],
                            n_rep: int,
                            rcParams: dict, **kwargs) -> pl.DataFrame:


    # Add imputation variables
    loc = rcParams.get("ion.aon_impute_loc", 1e4)
    scale = rcParams.get("ion.aon_impute_scale", 1e3)
    imp_df = pl.from_numpy(np.random.normal(loc=loc, scale=scale, size=(df.height, n_rep)), schema=["IMP"])
    df = df.hstack(imp_df)

    df = \
    df.with_columns(# Impute only on AON ions in control conditions
        pl.when(pl.col(f"{ctrl_name} ZC").eq(n_rep) & pl.col(f"{test_name} ZC").eq(0))
        .then(pl.col("IMP"))
        .otherwise(pl.concat_list(ctrl_ints))
        .alias(f"{ctrl_name} Intensity"),
        # Impute only on AON ions in test conditions
        pl.when(pl.col(f"{ctrl_name} ZC").eq(0) & pl.col(f"{test_name} ZC").eq(n_rep))
        .then(pl.col("IMP"))
        .otherwise(pl.concat_list(test_ints))
        .alias(f"{test_name} Intensity")
    ).with_columns(# Seperate out imputation (if perfromed) vars for control conditions
        pl.col(f"{ctrl_name} Intensity").list.get(i).alias(col) for i, col in enumerate(ctrl_ints)
    ).with_columns(# Seperate out imputation (if perfromed) vars for test conditions
        pl.col(f"{test_name} Intensity").list.get(i).alias(col) for i, col in enumerate(test_ints)
    ).with_columns(# Drop nulls for further computations, this column will be droped down-stream
        pl.col(f"{ctrl_name} Intensity").list.drop_nulls().alias(f"{ctrl_name} Intensity"),
        pl.col(f"{test_name} Intensity").list.drop_nulls().alias(f"{test_name} Intensity")
    ).drop(["IMP", f"{ctrl_name} ZC", f"{test_name} ZC"])

    return df


def _add_ttest(df: pl.DataFrame,
               ctrl_name: str,  test_name: str,
               n_rep: int,
               rcParams: dict, **kwargs) -> pl.DataFrame:

    df = \
    df.with_columns(# Prepare the descriptive stats for `.ttest_ind_from_stats()`
        pl.col(f"{ctrl_name} Intensity").list.mean().alias(f"{ctrl_name} Mean"),
        pl.col(f"{ctrl_name} Intensity").list.std().alias(f"{ctrl_name} Std"),

        pl.col(f"{test_name} Intensity").list.mean().alias(f"{test_name} Mean"),
        pl.col(f"{test_name} Intensity").list.std().alias(f"{test_name} Std"),
    ).with_columns(
        pl.struct(# Create struct with `.ttest_ind_from_stats()` var names
            mean1=pl.col(f"{ctrl_name} Mean"),
            std1=pl.col(f"{ctrl_name} Std"),
            nobs1=pl.col(f"{ctrl_name} Intensity").list.len(),
            mean2=pl.col(f"{test_name} Mean"),
            std2=pl.col(f"{test_name} Std"),
            nobs2=pl.col(f"{test_name} Intensity").list.len(),
            alternative=pl.col("Alternative Hypothesis")
        ).map_elements(# Perform T-test
            lambda x: sp.stats.ttest_ind_from_stats(**x, equal_var=False),
            return_dtype=pl.List(pl.Float64)
        ).alias("Stats")
    ).with_columns(# Seperate out T-test vars
        pl.col("Stats").list.first().alias("T-test"),
        pl.col("Stats").list.last().alias("P-value"),
    )

    return df


def _add_fdr(df: pl.DataFrame, rcParams: dict, **kwargs) -> pl.DataFrame:

    # Sort on P-value
    df = df.sort(by=["Protein ID", "P-value"], descending=[False, False])

    df = \
    df.hstack(
        df.group_by(by="Protein ID", maintain_order=True)
        .agg(# Calculate Adj. P-value
            pl.map_groups(
                exprs="P-value",
                function=lambda x: list(sp.stats.false_discovery_control(x[0])),
                return_dtype=pl.List(pl.Float64)
            ).alias("Adj. P-value"),
        )# Add Adj. P-value to the original `DataFrame`
        .explode(pl.col("Adj. P-value"))
        .select(pl.col("Adj. P-value"))
    )

    return df


def _add_ratio(df: pl.DataFrame,
               ctrl_name: str,  test_name: str,
               rcParams: dict, **kwargs) -> pl.DataFrame:

    df = \
    df.with_columns(# Calculate FC and CV
        (pl.col(f"{test_name} Mean") / pl.col(f"{ctrl_name} Mean")).alias("FC"),
        (pl.col(f"{test_name} Std") / pl.col(f"{test_name} Mean")).alias("CV")
    ).drop(# Cleaning up for output in ions
        [
            f"{ctrl_name} Mean", f"{ctrl_name} Std",
            f"{test_name} Mean", f"{test_name} Std",
            "Stats", f"{ctrl_name} Intensity", f"{test_name} Intensity"
        ]
    )

    return df


def _normalize_ratios(df: pl.DataFrame, trp_norm: pl.DataFrame, rcParams: dict, **kwargs) -> pl.DataFrame:

    trp_prot_fc = rcParams.get("trp_protein.fc_sig_tresh", 1.0)
    trp_prot_pval = rcParams.get("trp_protein.pval_sig_tresh", 0.01)

    trp_norm = trp_norm.select(pl.col("Protein ID", "P-value", "FC"))
    trp_norm = trp_norm.select(# Select only the significantly different proteins
        pl.col("Protein ID"),
        pl.when(pl.col("P-value").le(trp_prot_pval) & pl.col("FC").log(base=2).abs().ge(trp_prot_fc))
        .then(pl.col("FC"))
        .otherwise(pl.lit(0.0))
        .alias("Normalization Factor")
    )

    df = df.join(trp_norm, on="Protein ID", how="left")

    df = df.with_columns(# Apply normalization
        pl.col("Normalization Factor").fill_null(strategy="zero"),
        pl.when(pl.col("Normalization Factor").gt(0))
        .then(pl.col("FC") / pl.col("Normalization Factor"))
        .otherwise(pl.col("FC"))
        .alias("Normalized FC")
    )

    return df


def _add_start_end_aa(df: pl.DataFrame, rcParams: dict, **kwargs) -> pl.DataFrame:

    df = \
    df.with_columns(# Parse the Starting and Ending AA of the peptide
        pl.col("Peptide Sequence").str.head(1).alias("Start AA"),
        pl.col("Peptide Sequence").str.tail(1).alias("End AA"),
    )

    return df


def _add_half_trpytic(df: pl.DataFrame, rcParams: dict, **kwargs) -> pl.DataFrame:

    df = \
    df.with_columns(
        pl.when(# [K/R/-].~~~~~(K/R).[X]
            pl.col("Prev AA").is_in(["K", "R", "-"])
            & pl.col("End AA").is_in(["K", "R"])
            & ~pl.col("Next AA").is_in(["-"]))
        .then(pl.lit("FULL_TRP"))
        .when(# [K/R/-].~~~~~(X).[-]
            pl.col("Prev AA").is_in(["K", "R"])
            & pl.col("Next AA").is_in(["-"]))
        .then(pl.lit("FULL_TRP"))
        .when(# [M].2~~~~~(K/R).[X]
            pl.col("Prev AA").is_in(["M"])
            & pl.col("Start").eq(2)
            & pl.col("End AA").is_in(["K", "R"]))
        .then(pl.lit("FULL_TRP"))
        .when(# [K/R/-].~~~~~(X).[X]
            pl.col("Prev AA").is_in(["K", "R", "-"])
            & ~pl.col("End AA").is_in(["K", "R"])
            & ~pl.col("Next AA").is_in(["-"]))
        .then(pl.lit("C_SEMI"))
        .when(# [M].2~~~~~(X).[X]
            pl.col("Prev AA").is_in(["M"])
            & pl.col("Start").eq(2)
            & ~pl.col("End AA").is_in(["K", "R"]))
        .then(pl.lit("C_SEMI"))
        .when(# [X].~~~~~(K/R).[X]
            ~pl.col("Prev AA").is_in(["K", "R", "-"])
            & pl.col("End AA").is_in(["K", "R"])
            & ~pl.col("Next AA").is_in(["-"]))
        .then(pl.lit("N_SEMI"))
        .when(# [X].~~~~~(X).[-]
            ~pl.col("Prev AA").is_in(["K", "R", "-"])
            & pl.col("Next AA").is_in(["-"]))
        .then(pl.lit("N_SEMI"))
        .otherwise(pl.lit(None))
        .alias("Cleavage Type")
    ).filter(# Filter out over-digested peptides
        ~pl.col("Cleavage Type").is_null()
    ).with_columns(# Fill in `Half Tryptic`
        pl.col("Cleavage Type").ne("FULL_TRP").alias("Half Tryptic")
    )

    return df


def _add_cut_sites(df: pl.DataFrame, rcParams: dict, **kwargs) -> pl.DataFrame:

    df = \
    df.with_columns(# Adding unique identifier for all `Cleavage Type`
        pl.when(pl.col("Cleavage Type").eq("C_SEMI"))
        .then(pl.format("{}{}", pl.col("Next AA"), pl.col("End") + 1))
        .when(pl.col("Cleavage Type").eq("N_SEMI"))
        .then(pl.format("{}{}", pl.col("Start AA"), pl.col("Start")))
        .when((pl.col("Cleavage Type").eq("FULL_TRP")) & (pl.col("Prev AA").ne("-")))
        .then(
            pl.format(
                "{}{}-{}{}",
                pl.col("Prev AA"),
                pl.col("Start") - 1,
                pl.col("End AA"),
                pl.col("End"),
            )
        )
        .when((pl.col("Cleavage Type").eq("FULL_TRP")) & (pl.col("Prev AA").eq("-")))
        .then(
            pl.format(
                "{}{}-{}{}",
                pl.col("Start AA"),
                pl.col("Start"),
                pl.col("End AA"),
                pl.col("End"),
            )
        )
        .alias("Cut Site")
    ).with_columns(
        pl.format("{}_{}", pl.col("Protein ID"), pl.col("Cut Site")).alias("Cut Site ID")
    )

    return df


def _log2(df: pl.DataFrame, col: str) -> pl.DataFrame:

    df = df.with_columns(
        pl.when(pl.col(col).eq(0))
        .then(0.0)
        .otherwise(pl.col(col).log(base=2))
        .alias(f"Log2 {col}")
    )

    return df


def _log10(df: pl.DataFrame, col: str) -> pl.DataFrame:
    return df.with_columns(-pl.col(col).log10().alias(f"-Log10 {col}"))



import asyncio

import httpx
import nest_asyncio
import polars as pl
from Bio.SeqUtils.ProtParam import ProteinAnalysis

nest_asyncio.apply()

def n_batch(x: pl.Series, n: int):
    for i in range(0, len(x), n):
        yield x[i : i + n]


async def uniprot_api(acc: str) -> str:
    async with httpx.AsyncClient(
        base_url="https://rest.uniprot.org/uniprotkb", timeout=None
    ) as client:
        response = await client.get(f"{acc}.json")
        return response.text


async def uniprotkb(accessions: pl.Series) -> list:
    tasks = [uniprot_api(acc) for acc in accessions]
    return await asyncio.gather(*tasks)


# Parse JSON inside of this function
def uniprot_json(accessions: pl.Series):
    for acc in n_batch(accessions, 200):
        accessions_json = asyncio.run(uniprotkb(acc))


def bio_metadata(fasta: pl.DataFrame) -> pl.DataFrame:
    return fasta.with_columns(
        pl.col("Sequence").map_elements(lambda x: ProteinAnalysis(x), return_dtype=pl.Object)
        .alias("Analysis"),
    ).with_columns(
        pl.col("Analysis").map_elements(lambda x: x.isoelectric_point(), return_dtype=pl.Float64)
        .alias("pI"),
        pl.col("Analysis").map_elements(lambda x: x.molecular_weight(), return_dtype=pl.Float64)
        .alias("MW (Da)"),
        pl.col("Sequence").str.len_chars()
        .alias("Len (aa)"),
    ).drop(["Analysis", "Valid ID"])



rcParams: dict = {
    "ion.missing_intensity_thresh": 1,
    "ion.impute_type": "gaussian",
    "ion.aon_impute_loc": 1e4,
    "ion.aon_impute_scale": 1e3,
    "trp_protein.intensity_value": "MaxLFQ Intensity",
    "trp_protein.fc_sig_tresh": 1.0,
    "trp_protein.pval_sig_tresh": 0.01,
    "protein.fc_sig_thresh": 1.0,
    "protein.pval_sig_thresh": 0.01,
    "protein.adj_pval_sig_thresh": 0.05,
    "protein.high_fc_sig_thresh": 6.0,
    "protein.high_pval_sig_thresh": 0.016,
}

_LFQ_FP_FILES: list[str] = [
    "combined_ion.tsv",
    "combined_protein.tsv",
    "experiment_annotation.tsv",
]

_LFQ_FP_CONSTANT_ION_COLUMNS: list[str] = [
    "Peptide Sequence",
    "Modified Sequence",
    "Prev AA",
    "Next AA",
    "Start",
    "End",
    "Peptide Length",
    "M/Z",
    "Charge",
    "Compensation Voltage",
    "Assigned Modifications",
    "Protein",
    "Protein ID",
    "Entry Name",
    "Gene",
    "Protein Description",
    "Mapped Genes",
    "Mapped Proteins",
]

_LFQ_FP_VARIABLE_ION_COLUMNS: list[str] = [
    "Spectral Count",
    "Apex Retention Time",
    "Intensity",
    "Match Type",
]

_LFQ_FP_CONSTANT_PROTEIN_COLUMNS: list[str] = [
    "Protein",
    "Protein ID",
    "Entry Name",
    "Gene",
    "Protein Length",
    "Organism",
    "Protein Existence",
    "Description",
    "Protein Probability",
    "Top Peptide Probability",
    "Combined Total Peptides",
    "Combined Spectral Count",
    "Combined Unique Spectral Count",
    "Combined Total Spectral Count",
]

_LFQ_FP_VARIABLE_PROTEIN_COLUMNS: list[str] = [
    "Spectral Count",
    "Intensity",
    "MaxLFQ Intensity",
]

_FLIPPR_PROTEIN_COLUMNS: list[str] = [
    "Protein",
    "Protein ID",
    "Entry Name",
    "Gene",
    "Protein Length",
    "Organism",
    "Protein Existence",
    "Description",
    "Protein Probability",
]


_FLIPPR_PROTEIN_SUMMARY_COLUMNS: list[str] = [
    "Protein",
    "Entry Name",
    "Gene",
    "Protein Description",
    "Mapped Genes",
    "Mapped Proteins",
]

_FLIPPR_CUT_SITE_COLUMNS: list[str] = [
    "Cut Site",
    "Gene",
    "Entry Name",
    "Protein Description",
    "Mapped Genes",
    "Mapped Proteins",
    "Half Tryptic",
]

_FLIPPR_PEPTIDE_COLUMNS: list[str] = [
    "Prev AA",
    "Start AA",
    "End AA",
    "Next AA",
    "Start",
    "End",
    "Peptide Length",
    "Gene",
    "Entry Name",
    "Protein Description",
    "Mapped Genes",
    "Mapped Proteins",
    "Half Tryptic",
    "Cleavage Type",
]

_FLIPPR_MODIFIED_PEPTIDE_COLUMNS: list[str] = [
    "Prev AA",
    "Start AA",
    "End AA",
    "Next AA",
    "Start",
    "End",
    "Peptide Length",
    "Gene",
    "Entry Name",
    "Protein Description",
    "Mapped Genes",
    "Mapped Proteins",
    "Half Tryptic",
    "Cleavage Type",
]

_FLIPPR_COMBINE_KEY: dict[str, list] = {
    "CUT SITE": _FLIPPR_CUT_SITE_COLUMNS,
    "PEPTIDE": _FLIPPR_PEPTIDE_COLUMNS,
    "MODIFIED PEPTIDE": _FLIPPR_MODIFIED_PEPTIDE_COLUMNS,
}

# Thank you Holehouse lab!
_STANDARD_AA_CONVERSION = {
    "B": "N",
    "U": "C",
    "X": "G",
    "Z": "Q",
    "*": "",
    "-": "",
    " ": "",
}
